﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Grammars;
using Attribute = HeuristicLab.Grammars.Attribute;

namespace CodeGenerator {
  public class BruteForceCodeGen {

    private string usings = @"
using System.Collections.Generic;
using System.Linq;
using System;
";

    private string solverTemplate = @"
namespace ?PROBLEMNAME? {
  internal static class Grammar {
    ?GRAMMARCLASSCODE?
  }
  internal sealed class PartiallyProcessedSeq {
    private IEnumerable<string> remaining;
    private IEnumerable<int> path;
    public PartiallyProcessedSeq(IEnumerable<string> remaining) : this (Enumerable.Empty<int>(), remaining) {
    }
    public PartiallyProcessedSeq(IEnumerable<int> path, IEnumerable<string> remaining) {
      this.path = path;
      this.remaining = remaining;
    }

    public bool MoveNext() {
      if(!remaining.Any()) return false;
      else remaining = remaining.Skip(1);
      return true;
    }

    public string Current { 
      get {
        return remaining.FirstOrDefault();
      }
    }
    public IEnumerable<int> Path {
      get {
        return path;
      }
    }
    public IEnumerable<string> Remaining {
      get {
        return remaining;
      }
    }
    public override string ToString() {
      return path.Aggregate("""", (str, p) => str + "" "" + p) + "" >> "" +
             remaining.Aggregate("""", (str, r) => str + "" "" + r);
    }

  }
  public sealed class ?IDENT?Solver {

    public static void Main(string[] args) {
      var solver = new ?IDENT?Solver();
      solver.Start();
    }

    public ?IDENT?Solver() { 
      Initialize();
    }    

    private void Initialize() {
      ?INITCODE?
    }

    private bool IsBetter(double a, double b) { 
      return ?MAXIMIZATION? ? a > b : a < b;
    }

    #region path generator (BFS in the grammar to generate solutions)
    ?PATHGENERATORCODE?
    #endregion

    private IEnumerable<IEnumerable<int>> GeneratePaths() {
      var queue = new Queue<PartiallyProcessedSeq>();
      queue.Enqueue(new PartiallyProcessedSeq(new string[] { Grammar.RootSymbol }));

      while (queue.Count > 0) {
        var e = queue.Dequeue();
        var symb = e.Current; // the next symbol to generate paths for
        // Console.WriteLine(""Out: "" + e);

        var lastSymbol = !e.MoveNext();

        if (Grammar.IsTerminal(symb)) {
          // generate all paths for the terminal
          if (lastSymbol) {
            var pathsForSymb = GeneratePathsFor(symb);
            if (pathsForSymb.Any()) {
              foreach (var path in GeneratePathsFor(symb)) {
                yield return e.Path.Concat(path);
              }
            } else {
              yield return e.Path;
            }
          } else {
            var paths = GeneratePathsFor(symb);
            if(!paths.Any() || paths.Count() == 1) {
              yield return e.Path;
            } else {
              foreach (var path in paths) {
                var eAlt = new PartiallyProcessedSeq(e.Path.Concat(path), e.Remaining);
                queue.Enqueue(eAlt);
                // Console.WriteLine(""In: "" + eAlt);
              }
            }
          }
        } else {
          // non-terminal -> generate alternatives if necessary
          var alts = Grammar.GetAlternatives(symb);
          if (alts.Count() == 1) {
              var eAlt = new PartiallyProcessedSeq(e.Path, alts.Single().Concat(e.Remaining));
              queue.Enqueue(eAlt);
              // Console.WriteLine(""In: "" + eAlt);
          } else {
            int i = 0;
            foreach (var alt in alts) {
              var eAlt = new PartiallyProcessedSeq(e.Path.Concat(new int[] { i++ }), alt.Concat(e.Remaining));
              queue.Enqueue(eAlt);
              // Console.WriteLine(""In: "" + eAlt);
            }
          }
        }
      }
    }

    private void Start() {
      // generate possible paths through the grammar and evaluate each of them    
      var bestF = ?MAXIMIZATION? ? double.NegativeInfinity : double.PositiveInfinity;
      int n = 0;
      foreach(var path in GeneratePaths()) {
        var length = path.Count();
        currentPath = path;
        var f = Calculate();
        n++;
        if(IsBetter(f, bestF)) bestF = f;
        if (n%1000 == 0) Console.WriteLine(""{0}\t{1}\t{2}\t{3}"",n, length, bestF, f);
      }
    }

    private IEnumerable<int> currentPath;

    public double Calculate() {
      ?FITNESSFUNCTION?
    }

    ?ADDITIONALCODE?

    ?INTERPRETERSOURCE?

    ?CONSTRAINTSSOURCE?
  }
}";


    /// <summary>
    /// Generates the source code for a brute force searcher that can be compiled with a C# compiler
    /// </summary>
    /// <param name="ast">An abstract syntax tree for a GPDL file</param>
    public void Generate(GPDefNode ast) {
      var problemSourceCode = new StringBuilder();
      problemSourceCode.AppendLine(usings);

      GenerateProblem(ast, problemSourceCode);

      problemSourceCode.Replace("?PROBLEMNAME?", ast.Name);

      // write the source file to disk
      using (var stream = new StreamWriter(ast.Name + ".cs")) {
        stream.WriteLine(problemSourceCode.ToString());
      }
    }

    private void GenerateProblem(GPDefNode ast, StringBuilder problemSourceCode) {
      var grammar = CreateGrammarFromAst(ast);
      var problemClassCode =
        solverTemplate
          .Replace("?MAXIMIZATION?", ast.FitnessFunctionNode.Maximization.ToString().ToLowerInvariant())
          .Replace("?GRAMMARCLASSCODE?", GenerateGrammarClassCode(grammar))
          .Replace("?IDENT?", ast.Name)
          .Replace("?FITNESSFUNCTION?", ast.FitnessFunctionNode.SrcCode)
          .Replace("?INTERPRETERSOURCE?", GenerateInterpreterSource(grammar))
          .Replace("?INITCODE?", ast.InitCodeNode.SrcCode)
          .Replace("?ADDITIONALCODE?", ast.ClassCodeNode.SrcCode)
          .Replace("?CONSTRAINTSSOURCE?", GenerateConstraintMethods(ast.Terminals))
          .Replace("?PATHGENERATORCODE?", GeneratePathGeneratorCode(ast))
          ;

      problemSourceCode.AppendLine(problemClassCode).AppendLine();
    }

    private AttributedGrammar CreateGrammarFromAst(GPDefNode ast) {
      string startSymbolName = ast.Rules.First().NtSymbol;
      var startSymbolNode = ast.NonTerminals.Single(nt => nt.Ident == startSymbolName);
      // create startSymbol
      var g = new AttributedGrammar(new Symbol(startSymbolName, ParseSymbolAttributes(startSymbolNode.FormalParameters)));
      foreach (var rule in ast.Rules) {
        // create nt-symbol
        var ntSymbolName = rule.NtSymbol;
        var ntSymbolNode = ast.NonTerminals.Single(nt => nt.Ident == ntSymbolName);
        var attributes = ParseSymbolAttributes(ntSymbolNode.FormalParameters);
        var ntSymbol = new Symbol(ntSymbolName, attributes);
        foreach (var alt in GetAlternatives(rule.Alternatives)) {
          g.AddProductionRule(ntSymbol, alt);
        }
        // local initialization code
        if (!string.IsNullOrEmpty(rule.LocalCode)) g.AddLocalDefinitions(ntSymbol, rule.LocalCode);
      }
      return g;
    }

    private IEnumerable<IAttribute> ParseSymbolAttributes(string formalParameters) {
      return (from fieldDef in Util.ExtractParameters(formalParameters)
              select new Attribute(fieldDef.Identifier, fieldDef.Type, AttributeType.Parse(fieldDef.RefOrOut))).
        ToList();
    }

    private IEnumerable<Sequence> GetAlternatives(AlternativesNode altNode) {
      foreach (var alt in altNode.Alternatives) {
        yield return GetSequence(alt.Sequence);
      }
    }

    private Sequence GetSequence(IEnumerable<RuleExprNode> sequence) {
      Debug.Assert(sequence.All(s => s is CallSymbolNode || s is RuleActionNode));
      var l = new List<ISymbol>();
      foreach (var node in sequence) {
        var callSymbolNode = node as CallSymbolNode;
        var actionNode = node as RuleActionNode;
        if (callSymbolNode != null) {
          l.Add(new Symbol(callSymbolNode.Ident, ParseSymbolAttributes(callSymbolNode.ActualParameter)));
        } else if (actionNode != null) {
          l.Add(new SemanticSymbol("SEM", actionNode.SrcCode));
        }
      }
      return new Sequence(l);
    }


    private string GenerateGrammarClassCode(IGrammar grammar) {
      var sb = new StringBuilder();
      // RootSymbol
      sb.AppendFormat("public static string RootSymbol {{ get {{ return \"{0}\"; }} }}", grammar.StartSymbol.Name).AppendLine();
      sb.AppendLine("public static HashSet<string> terminalSymbols = new HashSet<string>() {");
      sb.AppendFormat("{0}", grammar.TerminalSymbols.Aggregate("", (str, symb) => str + "\"" + symb.Name + "\", "));
      sb.AppendLine("};");
      // GetAlternatives
      sb.AppendFormat("public static IEnumerable<IEnumerable<string>> GetAlternatives(string symbol) {{");
      sb.AppendFormat("switch(symbol) {{ ").AppendLine();
      foreach (var ntSymbol in grammar.NonTerminalSymbols) {
        sb.AppendFormat("case \"{0}\": {{ ", ntSymbol.Name).AppendLine();
        sb.Append("return new string[][] { ").AppendLine();
        foreach (var alt in grammar.GetAlternatives(ntSymbol)) {
          sb.Append("new string[] { ")
            .Append(alt.Skip(1).Aggregate("\"" + alt.First().Name + "\"", (str, symb) => str + ", \"" + symb.Name + "\""))
            .AppendLine("},");
        }
        sb.Append("};");
        sb.AppendLine("}");
      }
      sb.AppendLine(" default: { throw new InvalidOperationException(\"Unkown symbol: \"+symbol); }");
      sb.AppendLine("}");
      sb.AppendLine("}");

      sb.AppendLine("public static bool IsTerminal(string symbol) {");
      sb.AppendFormat("return terminalSymbols.Contains(symbol);");
      sb.AppendLine("}");
      // NumberOfAlternatives
      sb.AppendLine(
        "public static int NumberOfAlternatives(string symbol) { return GetAlternatives(symbol).Count(); }");
      return sb.ToString();
    }


    // produces helper methods for the attributes of all terminal nodes
    private string GenerateConstraintMethods(List<SymbolNode> symbols) {
      var sb = new StringBuilder();
      var terminals = symbols.OfType<TerminalNode>();
      foreach (var t in terminals) {
        sb.AppendLine(GenerateConstraintMethods(t));
      }
      return sb.ToString();
    }

    // generates helper methods for the attributes of a given terminal node
    private string GenerateConstraintMethods(TerminalNode t) {
      var sb = new StringBuilder();
      foreach (var c in t.Constraints) {
        var fieldType = t.FieldDefinitions.First(d => d.Identifier == c.Ident).Type;
        if (c.Type == ConstraintNodeType.Range) {
          sb.AppendFormat("public {0} GetMax{1}_{2}() {{ return {3}; }}", fieldType, t.Ident, c.Ident, c.RangeMaxExpression).AppendLine();
          sb.AppendFormat("public {0} GetMin{1}_{2}() {{ return {3}; }}", fieldType, t.Ident, c.Ident, c.RangeMinExpression).AppendLine();
        } else if (c.Type == ConstraintNodeType.Set) {
          sb.AppendFormat("public {0}[] GetAllowed{1}_{2}() {{ return {3}.ToArray(); }}", fieldType, t.Ident, c.Ident, c.SetExpression).AppendLine();
        }
        sb.AppendFormat("public {0} Get{1}_{2}Element(int i) {{ return GetAllowed{1}_{2}()[i]; }}", fieldType, t.Ident, c.Ident);
      }
      return sb.ToString();
    }


    // generates code for a breath-first-search generating all possible paths through the grammar
    private string GeneratePathGeneratorCode(GPDefNode ast) {
      var sb = new StringBuilder();
      foreach (var s in ast.NonTerminals.OfType<NonTerminalNode>()) {
        sb.Append(GeneratePathGeneratorCode(s));
      }

      sb.Append(GeneratePathGeneratorCode(ast.Terminals.OfType<TerminalNode>()));
      return sb.ToString();
    }

    private string GeneratePathGeneratorCode(NonTerminalNode s) {
      var sb = new StringBuilder();

      return sb.ToString();
    }

    // generates a method for the terminal that returns all an IEnumerable of int[] representing each possible combination of values for that terminal
    private string GeneratePathGeneratorCode(IEnumerable<TerminalNode> terminals) {
      var sb = new StringBuilder();
      sb.AppendLine("private IEnumerable<IEnumerable<int>> GeneratePathsFor(string tSymbol) {");
      sb.AppendLine("switch(tSymbol) {");
      foreach (var t in terminals) {
        sb.AppendFormat("case \"{0}\": {{", t.Ident).AppendLine();
        if (t.FieldDefinitions.Any()) {
          int i = 0;
          // open for loop for each field
          foreach (var f in t.FieldDefinitions) {
            sb.AppendFormat("for(int i{0}=0; i{0} < GetAllowed{1}_{2}().Length; i{0}++) {{", i++, t.Ident, f.Identifier)
              .
              AppendLine();
          }
          sb.AppendFormat("yield return new int[] {{ {0} }};",
                          Enumerable.Range(0, i).Select(ii => "i" + ii + ", ").Aggregate((str, e) => str + e));
          // close braces
          while (i-- > 0) {
            sb.AppendLine("}");
          }
          sb.AppendLine("break;");
        } else {
          sb.AppendLine("yield return Enumerable.Empty<int>();");
          sb.AppendLine("break;");
        }
        sb.AppendLine("}");
      }
      sb.AppendLine("} }");
      return sb.ToString();
    }

    private string GenerateInterpreterSource(AttributedGrammar grammar) {
      var sb = new StringBuilder();


      // find formal parameters of root node
      var attr = grammar.StartSymbol.Attributes;

      var formalParameter = grammar.StartSymbol.GetAttributeString();
      // actual parameter are the same as formalparameter only without type identifier
      string actualParameter;
      if (attr.Any())
        actualParameter = attr.Skip(1).Aggregate(attr.First().AttributeType + " " + attr.First().Name, (str, a) => str + ", " + a.AttributeType + " " + a.Name);
      else
        actualParameter = string.Empty;

      // generate entry method for evaluation. This is called from the min/max function
      // e.g.: ProgramRoot(ref int a0) { ProgramRoot(rootNode , ref a0); }
      sb.AppendFormat("void {0}({1}) {{", grammar.StartSymbol.Name, formalParameter).AppendLine();
      sb.AppendLine(" var path = currentPath.GetEnumerator();");
      sb.AppendFormat("{0}(path, {1});", grammar.StartSymbol.Name, actualParameter).AppendLine();
      sb.AppendLine("if(path.MoveNext()) throw new InvalidOperationException(); // assert that the full path has been processed");
      sb.AppendLine("}");

      // generate methods for all nonterminals and terminals using the grammar instance
      foreach (var s in grammar.NonTerminalSymbols) {
        sb.AppendLine(GenerateInterpreterMethod(grammar, s));
      }
      foreach (var s in grammar.TerminalSymbols) {
        sb.AppendLine(GenerateTerminalInterpreterMethod(s));
      }
      return sb.ToString();
    }

    private string GenerateTerminalInterpreterMethod(ISymbol s) {
      var sb = new StringBuilder();
      // if the terminal symbol has attributes then we must create values for these attributes
      if (!s.Attributes.Any())
        sb.AppendFormat("private void {0}(IEnumerator<int> path) {{", s.Name);
      else
        sb.AppendFormat("private void {0}(IEnumerator<int> path, {1}) {{", s.Name, s.GetAttributeString());

      // each field must match a formal parameter, assign a value for each parameter
      foreach (var element in s.Attributes) {
        // read next symbol
        sb.AppendLine("path.MoveNext();");
        sb.AppendFormat("{0} = Get{1}_{0}Element(path.Current)", element.Name, s.Name).AppendLine(";");
      }
      sb.AppendLine("}");
      return sb.ToString();
    }

    private string GenerateInterpreterMethod(AttributedGrammar g, ISymbol s) {
      var sb = new StringBuilder();
      if (!s.Attributes.Any())
        sb.AppendFormat("private void {0}(IEnumerator<int> path) {{", s.Name);
      else
        sb.AppendFormat("private void {0}(IEnumerator<int> path, {1}) {{", s.Name, s.GetAttributeString());
      // generate local definitions
      sb.AppendLine(g.GetLocalDefinitions(s));


      // if there are alternatives for this symbol -> choose alternative based on the path
      var alts = g.GetAlternatives(s);
      if (alts.Count() > 1) {
        // read next symbol
        sb.AppendLine("path.MoveNext();");
        int i = 0;
        sb.AppendLine("switch(path.Current) {");
        // generate a case for each alternative
        foreach (var l in alts) {
          sb.AppendFormat("case {0}: {{ ", i).AppendLine();
          foreach (var altS in g.GetAlternativeWithSemanticActions(s, i++)) {
            sb.AppendLine(GenerateSourceForAction(altS));
          }
          sb.AppendLine("break;").AppendLine("}");
        }
        sb.AppendLine("default: throw new System.InvalidOperationException();").AppendLine("}");
      } else {
        foreach (var altS in g.GetAlternativeWithSemanticActions(s, 0)) {
          sb.AppendLine(GenerateSourceForAction(altS));
        }
      }
      sb.AppendLine("}");
      return sb.ToString();
    }

    // helper for generating calls to other symbol methods
    private string GenerateSourceForAction(ISymbol s) {
      var action = s as SemanticSymbol;
      if (action != null) {
        return action.Code + ";";
      } else {
        if (!s.Attributes.Any())
          return string.Format("{0}(path);", s.Name);
        else
          return string.Format("{0}(path, {1});", s.Name, s.GetAttributeString());
      }
    }
  }
}
